import numpy as np

from pysketcher.arc import Arc
from pysketcher.point import Point


class Circle(Arc):
    def __init__(self, center: Point, radius: float, resolution=180):
        super().__init__(center, radius, 0, 2 * np.pi, resolution)



from typing import List

from pysketcher.curve import Curve
from pysketcher.point import Point
from pysketcher.style import Style


class Wall(Curve):
    _start: Point
    _end: Point
    _thickness: float

    def __init__(self, points: List[Point], thickness: float):
        self._start = points[0]
        self._end = points[-1]
        self._thickness = thickness

        def _displace(point: Point, point_before: Point, point_after: Point):
            # Displaces a point on our curve by thickness.
            # find a normal to the line between the point_before and the point_after
            # then displace by thickness from point in the direction of that normal
            return point + ((point_after - point_before).normal() * self._thickness)

        # at the start, there isn't a point_before, so use the start point
        new_points: List[Point] = [_displace(points[0], points[0], points[1])]

        for i in range(1, len(points) - 1):
            new_points += [_displace(points[i], points[i - 1], points[i + 1])]

        # and at the end there isn't a point_after, so use the end point
        new_points += [_displace(points[-1], points[-2], points[-1])]

        points = points + new_points[-1::-1]
        points += [self._start]

        super().__init__(points)
        self.style.fill_pattern = Style.FillPattern.CROSS

    def geometric_features(self):
        d = {"start": self._start, "end": self._end}
        return d



import numpy as np

from pysketcher.curve import Curve
from pysketcher.point import Point
from pysketcher.spline import Spline
from pysketcher.style import Style


class SketchyFunc1(Spline):
    """
    A typical function curve used to illustrate an "arbitrary" function.
    """

    domain = [1, 6]

    def __init__(self, name=None, name_pos="start", x_min=0, x_max=6, y_min=0, y_max=2):
        x = np.array([0, 2, 3, 4, 5, 6])
        y = np.array([1, 1.8, 1.2, 0.7, 0.8, 0.85])

        def scale_array(min: float, max: float, ps: np.array):
            return min - ps.min() + ps * (max - min) / (ps.max() - ps.min())

        xs = scale_array(x_min, x_max, x)
        ys = scale_array(y_min, y_max, y)
        points = Point.from_coordinate_lists(xs, ys)

        super().__init__(points)


class SketchyFunc2(Curve):
    """
    A typical function curve used to illustrate an "arbitrary" function.
    """

    domain = [0, 2.25]

    def __init__(self, x_min=0, x_max=2.25, y_min=0.046679703125, y_max=1.259375):
        a = 0
        b = 2.25
        resolution = 100
        xs = np.linspace(a, b, resolution + 1)
        ys = self(xs)
        # Scale x and y
        xs = x_min - xs.min() + xs * (x_max - x_min) / (xs.max() - xs.min())
        ys = y_min - ys.min() + ys * (y_max - y_min) / (ys.max() - ys.min())
        points = Point.from_coordinate_lists(xs, ys)

        super().__init__(points)

    def __call__(self, x):
        return 0.5 + x * (2 - x) * (0.9 - x)  # on [0, 2.25]


class SketchyFunc3(Spline):
    """
    A typical function curve used to illustrate an "arbitrary" function.
    """

    domain = [0, 6]

    def __init__(self, x_min=0, x_max=6, y_min=0.5, y_max=3.8):
        xs = np.array([0, 2, 3, 4, 5, 6])
        ys = np.array([0.5, 3.5, 3.8, 2, 2.5, 3.5])
        # Scale x and y
        xs = x_min - xs.min() + xs * (x_max - x_min) / (xs.max() - xs.min())
        ys = y_min - ys.min() + ys * (y_max - y_min) / (ys.max() - ys.min())
        points = Point.from_coordinate_lists(xs, ys)

        super().__init__(points)
        self.style.line_color = Style.Color.BLACK


class SketchyFunc4(Spline):
    """
    A typical function curve used to illustrate an "arbitrary" function.
    Can be a companion function to SketchyFunc3.
    """

    domain = [1, 6]

    def __init__(self, name_pos="start", x_min=0, x_max=6, y_min=0.5, y_max=1.8):
        xs = np.array([0, 2, 3, 4, 5, 6])
        ys = np.array([1.5, 1.3, 0.7, 0.5, 0.6, 0.8])
        # Scale x and y
        x = x_min - xs.min() + xs * (x_max - x_min) / (xs.max() - xs.min())
        y = y_min - ys.min() + ys * (y_max - y_min) / (ys.max() - ys.min())
        points = Point.from_coordinate_lists(xs, ys)

        super().__init__(points)



from pysketcher.angle import Angle
from pysketcher.point import Point
from pysketcher.shape import Shape
from pysketcher.style import Style, TextStyle


class Text(Shape):
    """
    Place `text` at the (x,y) point `position`, with the given
    fontsize (0 indicates that the default fontsize set in drawing_tool
    is to be used). The text is centered around `position` if `alignment` is
    'center'; if 'left', the text starts at `position`, and if
    'right', the right and of the text is located at `position`.
    """

    _style: TextStyle

    def __init__(self, text: str, position: Point, direction: Point = Point(1, 0)):
        super().__init__()
        self._text = text
        self._position = position
        self._direction = direction
        self._style = TextStyle()

    def rotate(self, angle: float, center: Point) -> "Text":
        direction = self._direction.rotate(angle, center)
        position = self._position.rotate(angle, center)
        return Text(self._text, position, direction)

    def __str__(self):
        return 'text "%s" at (%g,%g)' % (self._text, self._position.x, self._position.y)

    def __repr__(self):
        return repr(str(self))

    @property
    def style(self) -> TextStyle:
        return self._style

    @property
    def position(self) -> Point:
        return self._position

    @property
    def direction(self) -> Angle:
        return self._direction

    @property
    def text(self) -> str:
        return self._text

    @style.setter
    def style(self, text_style: TextStyle):
        self._style = text_style

    def set_alignment(self, alignment: TextStyle.Alignment) -> "Text":
        self.style.alignment = alignment
        return self

    def translate(self, vec: Point) -> "Text":
        new_text = Text(self.text, self.position, self.direction)
        new_text.style = self._style
        return new_text

    def scale(self, factor: float) -> "Text":
        raise NotImplementedError



import logging

import numpy as np


class Angle(np.float64):
    def __new__(cls, value: np.float64) -> "Angle":
        """A representation of a 2-Dimensional Angle

        Args:
            value: The value of the angle to create.

        Example:
            >>> Angle(1.0)
            Angle(1.0)
            >>> Angle(np.pi / 2)
            Angle(1.5707963267948966)
            >>> Angle(123)
            Angle(-2.6637061435917175)
            >>> Angle(2*np.pi)
            Angle(0.0)
        """
        value = cls._normalize(value)
        return super(cls, cls).__new__(cls, value)

    @staticmethod
    def _normalize(value: np.float64):
        logging.debug(f"Initial value: {value}")

        def order(x: np.float) -> np.float:
            return np.trunc(x / (2 * np.pi))

        if not (-np.pi < value <= np.pi):
            order_value = order(value)
            # logging.debug(f"Scaling by a factor of {order_value}")
            value = value - order_value * 2 * np.pi
            # logging.debug(f"Value after initial scaling: {value}")
            # TODO: for some reason, ``order`` sometimes overshoots. Need to fix this.
            while value <= -np.pi:
                # logging.debug("Nudging up")
                value = value + 2 * np.pi
            while value > np.pi:
                # logging.debug("Nudging down")
                value = value - 2 * np.pi
        # logging.debug(f"Final value: {value}")
        return value

    def __add__(self, other: np.float64) -> "Angle":
        res = super(self.__class__, self).__add__(other)
        return self.__class__(res)

    def __sub__(self, other: np.float64) -> "Angle":
        res = super(self.__class__, self).__sub__(other)
        return self.__class__(res)

    def __mul__(self, other: np.float64) -> "Angle":
        res = super(self.__class__, self).__mul__(other)
        return self.__class__(res)

    def __truediv__(self, other: np.float64) -> "Angle":
        res = super(self.__class__, self).__truediv__(other)
        return self.__class__(res)

    def __str__(self: np.float64) -> str:
        return f"{float(self)}"

    def __repr__(self: np.float64) -> str:
        return f"Angle({float(self)})"



from enum import Enum, auto, unique


@unique
class Color(Enum):
    RED = auto()
    GREEN = auto()
    BLUE = auto()
    CYAN = auto()
    MAGENTA = auto()
    YELLOW = auto()
    WHITE = auto()
    BLACK = auto()



from abc import ABC, abstractmethod

from pysketcher.drawable import Drawable


class Backend(ABC):
    pass

    @abstractmethod
    def add(self, shape: Drawable) -> None:
        pass

    @abstractmethod
    def erase(self) -> None:
        pass

    @abstractmethod
    def show(self) -> None:
        pass

    @abstractmethod
    def save(self, filename: str) -> None:
        pass



import numpy as np

from pysketcher.arc import ArcWithText
from pysketcher.style import Style


class Moment(ArcWithText):
    def __init__(
        self,
        text,
        center,
        radius,
        left=True,
        counter_clockwise=True,
        fontsize=0,
        text_spacing=1 / 60.0,
    ):
        style = Style.ArrowStyle.END if counter_clockwise else Style.ArrowStyle.START
        start_angle = np.pi / 2 if left else -np.pi / 2
        super().__init__(
            text,
            center,
            radius,
            start_angle,
            np.pi,
            text_spacing=text_spacing,
            resolution=180,
        )
        self.set_arrow(style)  # Curve object



from typing import List, Type, Union

import numpy as np

from pysketcher.backend.backend import Backend
from pysketcher.backend.matplotlib.matplotlib_backend import MatplotlibBackend
from pysketcher.drawable import Drawable


class Figure:
    _shapes: List[Drawable]
    _backend: Backend

    def __init__(
        self,
        x_min: np.float64,
        x_max: np.float64,
        y_min: np.float64,
        y_max: np.float64,
        backend: Type[Backend],
    ):
        self._backend = backend(x_min, x_max, y_min, y_max)
        self._shapes = []

    def add(self, shape: Drawable) -> None:
        self._shapes.append(shape)
        self._backend.add(shape)

    def show(self):
        self._backend.show()

    def save(self, filename: str) -> None:
        self._backend.save(filename)

    def erase(self):
        self._backend.erase()

    # def animate(
    #     self,
    #     drawing_tool: DrawingTool,
    #     time_points: List[float],
    #     action: Callable[["Shape", float, float], "Shape"],
    #     pause_per_frame: float = 0.5,
    #     dt: float = 0.5,
    #     title=None,
    # ):
    #
    #     for n, t in enumerate(time_points):
    #         drawing_tool.erase()
    #
    #         fig: Shape = action(self, t, dt)
    #         fig.draw(drawing_tool)



from typing import List

from pysketcher.arrow_with_text import ArrowWithText
from pysketcher.curve import Curve
from pysketcher.distance_with_text import DistanceWithText
from pysketcher.point import Point
from pysketcher.shape import Shape
from pysketcher.style import TextStyle


class Rectangle(Curve):
    """
    Rectangle specified by the point `lower_left_corner`, `width`,
    and `height`.
    """

    _lower_left_corner: Point
    _width: float
    _height: float

    def __init__(self, lower_left_corner: Point, width: float, height: float):

        self._width = width
        self._height = height
        self._lower_left_corner = lower_left_corner

        super().__init__(self._generate_points())

        # Dimensions
        dims = {
            "width": DistanceWithText(
                "width",
                self._lower_left_corner + Point(0, -height / 5.0),
                self._lower_left_corner + Point(width, -height / 5.0),
            ),
            "height": DistanceWithText(
                "height",
                self._lower_left_corner + Point(width + width / 5.0, 0),
                self._lower_left_corner + Point(width + width / 5.0, height),
            ),
            "lower_left_corner": ArrowWithText(
                "lower_left_corner",
                self._lower_left_corner - Point(width / 5.0, height / 5.0),
                self._lower_left_corner,
            ),
        }
        dims["height"]["text"].style.alignment = TextStyle.Alignment.LEFT
        self.dimensions = dims

    def _generate_points(self) -> List[Point]:
        return [
            self._lower_left_corner,
            self._lower_left_corner + Point(self._width, 0),
            self._lower_left_corner + Point(self._width, self._height),
            self._lower_left_corner + Point(0, self._height),
            self._lower_left_corner,
        ]

    def geometric_features(self):
        """
        Return dictionary with

        ==================== =============================================
        Attribute            Description
        ==================== =============================================
        lower_left           Lower left corner point.
        upper_left           Upper left corner point.
        lower_right          Lower right corner point.
        upper_right          Upper right corner point.
        lower_mid            Middle point on lower side.
        upper_mid            Middle point on upper side.
        center               Center point
        ==================== =============================================
        """
        d = {
            "lower_left": self._points[0],
            "lower_right": self._points[1],
            "upper_right": self._points[2],
            "upper_left": self._points[3],
        }
        d["lower_mid"] = (d["lower_left"] + d["lower_right"]) * 0.5
        d["upper_mid"] = (d["upper_left"] + d["upper_right"]) * 0.5
        d["left_mid"] = (d["lower_left"] + d["upper_left"]) * 0.5
        d["right_mid"] = (d["lower_right"] + d["upper_right"]) * 0.5
        d["center"] = Point(d["lower_mid"].x, d["left_mid"].y)
        return d

    def rotate(self, angle: float, center: Point) -> Curve:
        points = []
        for point in self._points:
            points.append(point.rotate(angle, center))
        return Curve(points)



from typing import Callable

from pysketcher.arrow import Arrow
from pysketcher.composition.composition import Composition
from pysketcher.line import Line
from pysketcher.point import Point
from pysketcher.spline import Spline


class VelocityProfile(Composition):
    _start: Point
    _height: float
    _profile: Callable[[float], Point]
    _num_arrows: int
    _scaling: float

    def __init__(
        self,
        start: Point,
        height: float,
        profile: Callable[[float], Point],
        num_arrows: int,
        scaling: float = 1,
    ):

        self._start = start
        self._height = height
        self._profile = profile
        self._num_arrows = num_arrows
        self._scaling = scaling

        shapes = dict()

        # Draw left line
        shapes["start line"] = Line(self._start, (self._start + Point(0, self._height)))

        # Draw velocity arrows
        dy = float(self._height) / (self._num_arrows - 1)

        end_points = []

        for i in range(self._num_arrows):
            start_position = Point(start.x, start.y + i * dy)
            end_position = start_position + profile(start_position.y) * self._scaling
            end_points += [end_position]
            shapes["arrow%d" % i] = Arrow(start_position, end_position)

        shapes["smooth curve"] = Spline(end_points)
        super().__init__(shapes)



from pysketcher.angle import Angle
from pysketcher.arrow import Arrow
from pysketcher.composition.composition import ShapeWithText
from pysketcher.point import Point
from pysketcher.text import Text


class Axis(ShapeWithText):
    def __init__(
        self,
        start: Point,
        length: float,
        label: str,
        rotation_angle: Angle = Angle(0.0),
        label_spacing=1.0 / 4.5,
    ):
        """
        Draw axis from start with `length` to the right
        (x axis). Place label at the end of the arrow tip.
        Then return `rotation_angle` (in degrees).
        The `label_spacing` denotes the space between the label
        and the arrow tip as a fraction of the length of the plot
        in x direction. A tuple can be given to adjust the position
        in both the x and y directions (with one parameter, the
        x position is adjusted).
        With `label_alignment` one can place
        the axis label text such that the arrow tip is to the 'left',
        'right', or 'center' with respect to the text field.
        The `label_spacing` and `label_alignment`parameters can
        be used to fine-tune the location of the label.
        """
        arrow = Arrow(start, start + Point(length, 0)).rotate(rotation_angle, start)
        # should increase spacing for downward pointing axis
        if type(label_spacing) != tuple:
            label_spacing = (label_spacing, label_spacing)
        label_pos = Point(
            start.x + length + label_spacing[0], start.y + label_spacing[1]
        )
        label = Text(label, label_pos).rotate(rotation_angle, start)

        super().__init__(arrow, label)

    def geometric_features(self):
        return self._shapes["arrow"].geometric_features()



from typing import List

import numpy as np

from pysketcher.angle import Angle
from pysketcher.point import Point
from pysketcher.shape import Shape


class Curve(Shape):
    """General curve as a sequence of (x,y) coordintes."""

    def __init__(self, points: List[Point]):
        """
        `x`, `y`: arrays holding the coordinates of the curve.
        """
        super().__init__()
        self._points = points
        # self.shapes must not be defined in this class
        # as self.shapes holds children objects:
        # Curve has no children (end leaf of self.shapes tree)

    @property
    def points(self):
        return self._points

    @property
    def xs(self):
        return np.array([p.x for p in self.points])

    @property
    def ys(self):
        return np.array([p.y for p in self.points])

    def rotate(self, angle: Angle, center: Point) -> "Curve":
        """
        Rotate all coordinates: `angle` is measured in radians
        center is the "origin" of the rotation.
        """
        ret_curve = Curve([p.rotate(angle, center) for p in self.points])
        ret_curve.style = self.style
        return ret_curve

    def scale(self, factor: np.float64) -> "Curve":
        """Scale all coordinates by `factor`: ``x = factor*x``, etc."""
        ret_curve = Curve(
            Point.from_coordinate_lists(factor * self.xs, factor * self.ys)
        )
        ret_curve.style = self.style
        return ret_curve

    def translate(self, vec: Point) -> "Curve":
        """Translate all coordinates by a vector `vec`."""
        ret_curve = Curve([p + vec for p in self.points])
        ret_curve.style = self.style
        return ret_curve

    def __str__(self):
        """Compact pretty print of a Curve object."""
        s = "%d (x,y) coords" % len(self._points)
        s += " %s" % self.style
        return s

    def __repr__(self):
        return str(self)



from matplotlib.axes import Axes

from pysketcher.backend.matplotlib.matplotlib_adapter import MatplotlibAdapter
from pysketcher.composition import Composition


class MatplotlibComposition(MatplotlibAdapter):

    _mplb: "MatplotlibBackend"

    def __init__(self, mplb: "MatplotlibBackend"):
        self._mplb = mplb

    def plot(self, shape: Composition, axes: Axes):
        shape.apply(self._mplb.add)



from enum import Enum, auto

from pysketcher.arrow import DoubleArrow
from pysketcher.arrow_with_text import Text
from pysketcher.composition.composition import ShapeWithText
from pysketcher.point import Point


class DistanceWithText(ShapeWithText):
    """
    Arrow <-> with text (usually a symbol) at the midpoint, used for
    identifying a some distance in a figure.  The text is placed
    slightly to the right of vertical-like arrows, with text displaced
    `text_spacing` times to total distance in x direction of the plot
    area. The text is by default aligned 'left' in this case. For
    horizontal-like arrows, the text is placed the same distance
    above, but aligned 'center' by default (when `alignment` is None).
    """

    class TextPosition(Enum):
        START = auto()
        MIDDLE = auto()
        END = auto()

        def swap(self) -> "TextPosition":
            swap_dict = {
                self.START: self.END,
                self.END: self.START,
                self.MIDDLE: self.MIDDLE,
            }
            return swap_dict[self]

    def __init__(
        self,
        text: str,
        start: Point,
        end: Point,
        spacing: float = 1 / 3.0,
        text_position: TextPosition = TextPosition.MIDDLE,
    ):
        self._start = start
        self._end = end

        # Decide first if we have a vertical or horizontal arrow
        vertical = abs(end.x - start.x) < 2 * abs(end.y - start.y)

        if vertical:
            # Force end above start
            if end.y < start.y:
                start, end = end, start
                text_position = text_position.swap()
        else:  # horizontal arrow
            # Force start to the right of end
            if start.x < end.x:
                start, end = end, start
                text_position = text_position.swap()

        normal = (end - start).normal()

        if text_position == self.TextPosition.MIDDLE:
            # if the text is in the middle of the line, then place it
            # at a point spacing away normal to the midpoint
            mid = (start + end) * 0.5  # midpoint of start-end line
            text_point = mid - normal * spacing
        elif text_position == self.TextPosition.START:
            # if the text is at the start of the line, then place it
            # at a point spacing away normal to the start
            text_point = start - normal * spacing
        elif text_position == self.TextPosition.END:
            # if the text is at the start of the line, then place it
            # at a point spacing away normal to the end
            text_point = end - normal * spacing
        else:
            raise ValueError(
                "The value of `text_position`: %s is not valid", text_position
            )

        text = Text(text, text_point)

        arrow = DoubleArrow(start, end)
        arrow.line_color = "black"
        arrow.line_width = 1
        super().__init__(arrow, text)

    def geometric_features(self):
        d = self._shapes["arrow"].geometric_features()
        d["text_position"] = self._shapes["text"].position
        return d



import numpy as np

from pysketcher import Curve, Point
from pysketcher.composition.composition import Composition
from pysketcher.line import Line


class Spring(Composition):
    """
    Specify a *vertical* spring, starting at `start` and with `length`
    as total vertical length. In the middle of the spring there are
    `num_windings` circular windings to illustrate the spring. If
    `teeth` is true, the spring windings look like saw teeth,
    otherwise the windings are smooth circles.  The parameters `width`
    (total width of spring) and `bar_length` (length of first and last
    bar are given sensible default values if they are not specified
    (these parameters can later be extracted as attributes, see table
    below).

    Examples:
        >>> L = 12.
        >>> H = L / 6.
        >>> start = ps.Point(0., 0.)
        >>> s = ps.Spring(start, L)
        >>> fig = ps.Figure(-2, 2, -1, L+H,backend=MatplotlibBackend)
        >>> fig.add(s)
        >>> fig.show()
    """

    spring_fraction = 1.0 / 2  # fraction of total length occupied by spring

    def __init__(
        self,
        start: Point,
        length: float,
        width: float = None,
        bar_length: float = None,
        num_windings: int = 11,
        teeth: bool = False,
    ):
        B = start
        n = num_windings - 1  # n counts teeth intervals
        if n <= 6:
            n = 7
        # n must be odd:
        if n % 2 == 0:
            n = n + 1
        L = length
        if width is None:
            w = L / 10.0
        else:
            w = width / 2.0
        s = bar_length

        shapes = {}
        if s is None:
            f = Spring.spring_fraction
            s = L * (1 - f) / 2.0  # start of spring

        self.bar_length = s  # record
        self.width = 2 * w

        p0 = Point(B.x, B.y + s)
        p1 = Point(B.x, B.y + L - s)
        p2 = Point(B.x, B.y + L)

        if s >= L:
            raise ValueError(
                "length of first bar: %g is larger than total length: %g" % (s, L)
            )

        shapes["bar1"] = Line(B, p0)
        spring_length = L - 2 * s
        t = spring_length / n  # height increment per winding
        if teeth:
            resolution = 4
        else:
            resolution = 90
        q = np.linspace(0, n, n * resolution + 1)
        xs = p0.x + w * np.sin(2 * np.pi * q)
        ys = p0.y + q * t
        points = Point.from_coordinate_lists(xs, ys)
        shapes["spiral"] = Curve(points)
        shapes["bar2"] = Line(p1, p2)
        super().__init__(shapes)



from enum import Enum, auto, unique
from typing import Tuple, Union

from pysketcher.arrow import Arrow
from pysketcher.composition import ShapeWithText
from pysketcher.point import Point
from pysketcher.text import Text


class ArrowWithText(ShapeWithText):
    """
    As class Text, but an arrow is drawn from the mid part of the text
    to some point `arrow_tip`.
    """

    _DEFAULT_SPACING: Point = Point(0.15, 0.15)

    @unique
    class TextPosition(Enum):
        START = auto()
        END = auto()

    def __init__(
        self,
        text: str,
        start: Point,
        end: Point,
        text_position: TextPosition = TextPosition.START,
        spacing: Union[float, Point] = None,
    ):

        spacing = spacing if spacing else self._DEFAULT_SPACING
        if not issubclass(spacing.__class__, Point):
            spacing = Point(spacing, spacing)

        if text_position == self.TextPosition.START:
            text = Text(text, start + spacing)
        if text_position == self.TextPosition.END:
            text = Text(text, end + spacing)

        arrow = Arrow(start, end)
        super().__init__(arrow, text)



from collections import namedtuple
from enum import Enum, auto, unique

Arrow = namedtuple("Arrow", "start end")


class Style:
    @unique
    class LineStyle(Enum):
        SOLID = auto()
        DOTTED = auto()
        DASH_DOT = auto()
        DASHED = auto()

    @unique
    class Color(Enum):
        GREY = auto()
        BLACK = auto()
        BROWN = auto()
        RED = auto()
        ORANGE = auto()
        YELLOW = auto()
        GREEN = auto()
        CYAN = auto()
        BLUE = auto()
        MAGENTA = auto()
        PURPLE = auto()
        WHITE = auto()

    @unique
    class FillPattern(Enum):
        VERTICAL = auto()
        HORIZONTAL = auto()
        CROSS = auto()
        SQUARE = auto()
        STAR = auto()
        DOT = auto()
        CIRCLE = auto()
        SMALL_CIRCLE = auto()
        UP_RIGHT_TO_LEFT = auto()
        UP_LEFT_TO_RIGHT = auto()

    @unique
    class ArrowStyle(Enum):
        START = Arrow(True, False)
        END = Arrow(False, True)
        DOUBLE = Arrow(True, True)

    _line_style: LineStyle
    _line_width: float
    _line_color: Color
    _fill_pattern: FillPattern
    _fill_color: Color
    _arrow: ArrowStyle
    _shadow: int

    def __init__(self):
        self._line_style = self.LineStyle.SOLID
        self._line_width = 1.0
        self._line_color = self.Color.BLACK
        self._fill_pattern = None
        self._fill_color = None
        self._arrow = None
        self._shadow = None

    @property
    def line_style(self) -> LineStyle:
        return self._line_style

    @line_style.setter
    def line_style(self, line_style: LineStyle):
        self._line_style = line_style

    @property
    def line_width(self) -> float:
        return self._line_width

    @line_width.setter
    def line_width(self, line_width: float):
        self._line_width = line_width

    @property
    def line_color(self) -> Color:
        return self._line_color

    @line_color.setter
    def line_color(self, color: Color):
        self._line_color = color

    @property
    def fill_color(self) -> Color:
        return self._fill_color

    @fill_color.setter
    def fill_color(self, color: Color):
        self._fill_color = color

    @property
    def fill_pattern(self) -> FillPattern:
        return self._fill_pattern

    @fill_pattern.setter
    def fill_pattern(self, fill_pattern: FillPattern):
        self._fill_pattern = fill_pattern
        if not self.fill_color:
            self.fill_color = self.Color.WHITE

    @property
    def arrow(self) -> ArrowStyle:
        return self._arrow

    @arrow.setter
    def arrow(self, arrow: ArrowStyle):
        self._arrow = arrow

    @property
    def shadow(self) -> float:
        return self._shadow

    @shadow.setter
    def shadow(self, shadow: float):
        self._shadow = shadow

    def __str__(self):
        return (
            "line_style: %s, line_width: %s, line_color: %s,"
            " fill_pattern: %s, fill_color: %s, arrow: %s shadow: %s"
            % (
                self.line_style,
                self.line_width,
                self.line_color,
                self.fill_pattern,
                self.fill_color,
                self.arrow,
                self.shadow,
            )
        )


class TextStyle(Style):
    class FontFamily(Enum):
        SERIF = auto()
        SANS = auto()
        MONO = auto()

    class Alignment(Enum):
        LEFT = auto()
        CENTER = auto()
        RIGHT = auto()

    _font_size: float
    _font_family: FontFamily
    _alignment: Alignment

    def __init__(self,):
        super().__init__()
        self._font_size = 12
        self._font_family = self.FontFamily.SANS
        self._alignment = self.Alignment.CENTER

    @property
    def font_size(self) -> float:
        return self._font_size

    @font_size.setter
    def font_size(self, font_size: float):
        self._font_size = font_size

    @property
    def font_family(self) -> FontFamily:
        return self._font_family

    @font_family.setter
    def font_family(self, font_family: FontFamily):
        self._font_family = font_family

    @property
    def alignment(self):
        return self._alignment

    @alignment.setter
    def alignment(self, alignment: Alignment):
        self._alignment = alignment



import logging
import warnings
from typing import List, Tuple

import numpy as np

from pysketcher.angle import Angle
from pysketcher.warning import LossOfPrecisionWarning


class Point:
    """
    Simple Point class which implements basic point arithmetic.
    """

    _x: np.float64
    _y: np.float64

    def __init__(self, x: np.float64, y: np.float64):
        """
        Parameters
        ---------
        x: float
            The x co-ordinate
        y: float
            The y co-ordinate
        """
        self._x = np.float64(x)
        self._y = np.float64(y)

    def __add__(self, other: "Point") -> "Point":
        return Point(self.x + other._x, self.y + other.y)

    def __sub__(self, other: "Point") -> "Point":
        return Point(self.x - other._x, self.y - other.y)

    def __mul__(self, scalar: np.float64) -> "Point":
        return Point(self.x * scalar, self.y * scalar)

    def __abs__(self) -> np.float64:
        return np.hypot(self.x, self.y)

    def __eq__(self, other: "Point") -> bool:
        return self._isclose(self.x, other.x) and self._isclose(self.y, other.y)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "P(%s, %s)" % (
            np.format_float_scientific(self.x),
            np.format_float_scientific(self.y),
        )

    @staticmethod
    def _isclose(x: np.float64, y: np.float64) -> bool:
        return np.isclose(x, y, atol=1e-4)

    @property
    def x(self) -> np.float64:
        return self._x

    @property
    def y(self) -> np.float64:
        return self._y

    def unit_vector(self) -> "Point":
        if self._isclose(abs(self), 0.0):
            raise ZeroDivisionError("Length of Vector cannot be Zero")
        return self * (1 / (abs(self)))

    def angle(self) -> Angle:
        angle = Angle(np.arctan2(self.y, self.x))
        return angle

    def radius(self) -> np.float64:
        return abs(self)

    def normal(self) -> "Point":
        uv = self.unit_vector()
        return Point(-uv.y, uv.x)

    def rotate(self, angle: Angle, center: "Point") -> "Point":
        """Rotate point an `angle` (in radians) around (`x`,`y`)."""

        if not type(angle) == Angle:
            angle = Angle(angle)

        # Check for a few degenerate cases:
        if angle == Angle(0.0):
            p = self
        elif angle == Angle(np.pi / 2):
            p = Point(center.x - self.y + center.y, center.y + self.x - center.x)
        elif angle == Angle(np.pi):
            p = Point(2 * center.x - self.x, 2 * center.y - self.y)
        elif angle == Angle(-np.pi / 2):
            p = Point(center.x + self.y - center.y, center.y - self.x + center.x)
        else:
            c = np.cos(angle)
            s = np.sin(angle)
            p = Point(
                center.x + (self.x - center.x) * c - (self.y - center.y) * s,
                center.y + (self.x - center.x) * s + (self.y - center.y) * c,
            )
        return p

    def scale(self, factor: np.float64) -> "Point":
        """Scale point coordinates by `factor`: ``x = factor*x``, etc."""
        return self * factor

    def translate(self, vec: "Point") -> "Point":
        """Translate point by a vector `vec`."""
        return self + vec

    @staticmethod
    def from_coordinate_lists(
        xs: List[np.float64], ys: List[np.float64]
    ) -> List["Point"]:
        if len(xs) != len(ys):
            raise ValueError("xs and ys must be the same length")
        return [Point(xs[i], ys[i]) for i in range(len(xs))]

    @staticmethod
    def to_coordinate_lists(
        ps: List["Point"]
    ) -> Tuple[List[np.float64], List[np.float64]]:
        xs = [p.x for p in ps]
        ys = [p.y for p in ps]
        return xs, ys



import warnings
from copy import copy
from typing import Tuple

import numpy as np

from pysketcher.curve import Curve
from pysketcher.point import Point


class Line(Curve):

    _start: Point
    _end: Point
    _a: np.float64
    _b: np.float64
    _c: np.float64
    _d: np.float64
    _vertical: bool
    _horizontal: bool

    def __init__(self, start: Point, end: Point):
        """ A representation of a line primitive.

        Args:
            start: The starting point of the line.
            end: The end point of the line.

        Example:
            >>> a = Line(Point(1.,2.), Point(4.,3.))
            >>> b = a.rotate(np.pi / 2, Point(1.,2.))
        """
        if start == end:
            raise ValueError("Cannot specify a line with two equal points.")
        self._start = start
        self._end = end
        self._horizontal = False
        self._vertical = False
        self._a = self._b = self._c = self._d = None
        super().__init__([self._start, self._end])
        self._compute_formulas()

    def _compute_formulas(self):
        # Define equations for line:
        # y = a*x + b,  x = c*y + d
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "divide by zero encountered in double_scalars"
            )
            warnings.filterwarnings(
                "ignore", "invalid value encountered in double_scalars"
            )
            warnings.filterwarnings("ignore", "overflow encountered in double_scalars")
            self._a = (self._end.y - self._start.y) / (self._end.x - self._start.x)
            self._b = self._start.y - self._a * self._start.x
            if np.isnan(self._a) or np.isinf(self._a):
                # Vertical line, y is not a function of x
                self._vertical = True
                self._c = 0.0
                self._d = self._end.x
            else:
                self._c = 1.0 / self._a
                self._d = self._b / self._a
                if np.isnan(self._c) or np.isinf(self._c):
                    self._horizontal = True

    @property
    def start(self):
        """The starting point of the line."""
        return self._start

    @property
    def end(self):
        """The end point of the line."""
        return self._end

    def __call__(self, x: np.float64 = None, y: np.float64 = None):
        """Given x, return y on the line, or given y, return x.

        Args:
            x: the value of x for which a value of y should be calculated
            y: the value of y for which a value of x should be calculated

        Returns:
            the appropriate value of either y or x.

        Raises:
            ValueError: If the line is horizontal and y is provided, or if
            the line is vertical and x is provided.
        """
        self._compute_formulas()
        if self._horizontal and y:
            raise ValueError("Value of x is not dependent on the value of y.")
        if self._vertical and x:
            raise ValueError("Value of y is not dependent on the value of x.")
        return self._a * x + self._b if x else self._c * y + self._d

    def interval(
        self,
        x_range: Tuple[np.float64, np.float64] = None,
        y_range: Tuple[np.float64, np.float64] = None,
    ):
        """Returns a smaller portion of a line.

        Args:
            x_range: The range of x-coordinates which should be used to obtain the segment
            y_range: The range of y-coordinates which should be used to obtain the segment

        """
        if x_range and y_range:
            raise ValueError("Cannot specify both x_range and y_range.")
        if x_range is not None:
            return Line(
                Point(x_range[0], self(x_range[0])), Point(x_range[1], self(x_range[1]))
            )
        elif y_range is not None:
            return Line(
                Point(self(y_range[0]), y_range[0]), Point(self(y_range[1]), y_range[1])
            )

    def rotate(self, angle: np.float64, center: Point) -> "Line":
        """Returns a copy of a line rotated through an angle about a point.

        Args:
            angle:
            center:
        """
        start = self._start.rotate(angle, center)
        end = self._end.rotate(angle, center)
        line = Line(start, end)
        line.style = copy(self.style)
        return line



from pysketcher.arrow import Arrow
from pysketcher.composition.composition import Composition
from pysketcher.point import Point
from pysketcher.rectangle import Rectangle


class UniformLoad(Composition):
    """
    Downward-pointing arrows indicating a vertical load.
    The arrows are of equal length and filling a rectangle
    specified as in the :class:`Rectangle` class.

    Recorded geometric features:

    ==================== =============================================
    Attribute            Description
    ==================== =============================================
    mid_top              Middle point at the top of the row of
                         arrows (often used for positioning a text).
    ==================== =============================================
    """

    def __init__(
        self, lower_left_corner: Point, width: float, height: float, num_arrows=10
    ):
        box = Rectangle(lower_left_corner, width, height)
        shapes = {"box": box}
        dx = float(width) / (num_arrows - 1)
        for i in range(num_arrows):
            x = lower_left_corner.x + i * dx
            start = Point(x, lower_left_corner.y + height)
            end = Point(x, lower_left_corner.y)
            shapes["arrow%d" % i] = Arrow(start, end)
        super().__init__(shapes)

    def geometric_features(self):
        return {"mid_top": self["box"].geometric_features()["upper_mid"]}



from pysketcher.composition.composition import Composition
from pysketcher.distance_with_text import DistanceWithText
from pysketcher.point import Point
from pysketcher.rectangle import Rectangle
from pysketcher.text import Text
from pysketcher.triangle import Triangle


class SimpleSupport(Composition):
    _position: Point
    _size: float

    def __init__(self, position: Point, size: float):
        p0 = Point(position.x - size / 2.0, position.y - size)
        p1 = Point(position.x + size / 2.0, position.y - size)
        triangle = Triangle(p0, p1, position)
        gap = size / 5.0
        h = size / 4.0  # height of rectangle
        p2 = Point(p0.x, p0.y - gap - h)
        rectangle = Rectangle(p2, size, h)
        shapes = {"triangle": triangle, "rectangle": rectangle}
        super().__init__(shapes)

        self._dimensions = {
            "position": Text("position", position),
            "size": DistanceWithText(
                "size", Point(p2.x, p2.y - size), Point(p2.x + size, p2.y - size)
            ),
        }

    def geometric_features(self):
        t = self["triangle"]
        r = self["rectangle"]
        d = {
            "pos": t.geometric_features()["p2"],
            "mid_support": r.geometric_features()["lower_mid"],
        }
        return d



import logging

import numpy as np

from pysketcher.angle import Angle
from pysketcher.composition.composition import ShapeWithText
from pysketcher.curve import Curve
from pysketcher.point import Point
from pysketcher.text import Text


class Arc(Curve):
    """
    A representation of a continuous connected subset of a circle.
    """

    _center: Point
    _radius: float
    _start_angle: Angle
    _arc_angle: Angle
    _resolution: int

    def __init__(
        self,
        center: Point,
        radius: float,
        start_angle: Angle,
        arc_angle: Angle,
        resolution: int = 180,
    ):
        # Must record some parameters for __call__
        self._center = center
        self._radius = radius
        self._start_angle = Angle(start_angle)
        self._arc_angle = Angle(arc_angle)
        self._resolution = resolution

        if self._arc_angle == 0.0:
            # assume a full circle
            ts = np.linspace(0.0, 2.0 * np.pi, resolution + 1)
        else:
            ts = np.linspace(0.0, self._arc_angle, resolution + 1)

        points = [self(t) for t in ts]
        super().__init__(points)

    def __call__(self, theta: Angle) -> Point:
        """
        Return (x,y) point at start_angle + theta.
        """
        if self._arc_angle != 0.0 and theta > self._arc_angle:
            raise ValueError("Theta is outside the bounds of the arc")
        iota = Angle(self.start_angle + theta)
        ret_point = Point(
            self.center.x + self.radius * np.cos(iota),
            self.center.y + self.radius * np.sin(iota),
        )
        return ret_point

    @property
    def start_angle(self) -> Angle:
        return self._start_angle

    @property
    def arc_angle(self) -> Angle:
        return self._arc_angle

    @property
    def end_angle(self) -> Angle:
        return self._start_angle + self._arc_angle

    @property
    def radius(self) -> float:
        return self._radius

    @property
    def center(self) -> Point:
        return self._center

    @property
    def start(self) -> Point:
        return self(0.0)

    @property
    def end(self) -> Point:
        return self(self.arc_angle)

    def translate(self, vec: Point) -> "Arc":
        arc = Arc(
            self._center + vec,
            self._radius,
            self._start_angle,
            self._arc_angle,
            self._resolution,
        )
        arc.style = self.style
        return arc


class ArcWithText(ShapeWithText):
    def __init__(
        self,
        text: str,
        center: Point,
        radius: float,
        start_angle: Angle,
        arc_angle: Angle,
        resolution: int = 180,
        text_spacing: float = 1 / 6.0,
    ):
        arc = Arc(center, radius, start_angle, arc_angle, resolution)
        mid = arc(arc_angle / 2.0)
        normal = (mid - center).unit_vector()
        text_pos = mid - (normal * text_spacing)
        text = Text(text, text_pos)
        super().__init__(arc, text)



import matplotlib.pyplot as plt

from pysketcher.backend.matplotlib.matplotlib_adapter import MatplotlibAdapter
from pysketcher.backend.matplotlib.matplotlib_style import MatplotlibTextStyle
from pysketcher.text import Text


class MatplotlibText(MatplotlibAdapter):
    def plot(self, text: Text, axes: plt.Axes):
        """
        Write `text` string at a position (centered, left, right - according
        to the `alignment` string). `position` is a point in the coordinate
        system.
        If ``arrow+tip != None``, an arrow is drawn from the text to a point
        (on a curve, for instance). The arrow_tip argument is then
        the (x,y) coordinates for the arrow tip.
        fontsize=0 indicates use of the default font as set by
        ``set_fontsize``.
        """
        mpl_style = MatplotlibTextStyle(text.style)
        kwargs = {}
        if mpl_style.font_family is not None:
            kwargs["family"] = mpl_style.font_family
        if mpl_style.fill_color is not None:
            kwargs["backgroundcolor"] = mpl_style.fill_color
        if mpl_style.line_color is not None:
            kwargs["color"] = mpl_style.line_color

        rotation_angle = text.direction.angle()
        if rotation_angle != 0.0:
            kwargs["rotation"] = rotation_angle

        axes.text(
            text.position.x,
            text.position.y,
            text.text,
            horizontalalignment=mpl_style.alignment,
            fontsize=mpl_style.font_size,
            **kwargs
        )



import logging
from typing import Dict, Type

import matplotlib.pyplot as plt

from pysketcher.backend.backend import Backend
from pysketcher.backend.matplotlib.matplotlib_adapter import MatplotlibAdapter
from pysketcher.backend.matplotlib.matplotlib_composition import MatplotlibComposition
from pysketcher.backend.matplotlib.matplotlib_curve import MatplotlibCurve
from pysketcher.backend.matplotlib.matplotlib_text import MatplotlibText
from pysketcher.composition import Composition
from pysketcher.curve import Curve
from pysketcher.drawable import Drawable
from pysketcher.text import Text

plt.rc("text", usetex=True)
plt.rcParams["text.latex.preamble"] = r"\usepackage{amsmath}"


class MatplotlibBackend(Backend):
    """
    Simple interface for plotting. This interface makes use of
    Matplotlib for plotting.

    """

    _fig: plt.Figure
    _axes: plt.Axes
    _x_min: float
    _y_min: float
    _x_max: float
    _y_max: float

    def __init__(self, x_min, x_max, y_min, y_max):
        plt.ion()
        self._x_min = x_min
        self._x_max = x_max
        self._y_min = y_min
        self._y_max = y_max
        self._fig = plt.figure(
            figsize=[x_max - x_min, y_max - y_min], tight_layout=False
        )
        self._axes = self._fig.gca()
        self._configure_axes()

    def _configure_axes(self):
        self._axes.set_xlim(self._x_min, self._x_max)
        self._axes.set_ylim(self._y_min, self._y_max)
        self._axes.set_axis_off()

    def add(self, shape: Drawable) -> None:
        for typ, adapter in self._adapters.items():
            if issubclass(shape.__class__, typ):
                adapter.plot(shape, self._axes)

    def erase(self):
        self._fig.clear()
        self._axes = self._fig.gca()
        self._configure_axes()

    def show(self):
        self._fig.sca(self._axes)
        self._fig.canvas.draw()
        self._fig.show()

    def save(self, filename: str) -> None:
        logging.info(f"Saving to {filename}.")
        # TODO: manage formats
        self._fig.savefig(filename)

    @property
    def _adapters(self) -> Dict[Type, MatplotlibAdapter]:
        return {
            Curve: MatplotlibCurve(),
            Text: MatplotlibText(),
            Composition: MatplotlibComposition(self),
        }



from abc import ABC, abstractmethod
from typing import Any

from matplotlib.axes import Axes

from pysketcher.drawable import Drawable


class MatplotlibAdapter(ABC):
    @staticmethod
    @abstractmethod
    def plot(shape: Drawable, axes: Axes):
        pass



from typing import List

import matplotlib.pyplot as plt
from matplotlib import transforms

import pysketcher as ps
from pysketcher.backend.matplotlib.matplotlib_adapter import MatplotlibAdapter
from pysketcher.backend.matplotlib.matplotlib_style import MatplotlibStyle
from pysketcher.style import Style


class MatplotlibCurve(MatplotlibAdapter):
    types: List[type] = [ps.Curve]

    def plot(self, curve: ps.Curve, axes: plt.Axes) -> None:
        """Draw a curve with coordinates x and y (arrays)."""
        mpl_style = MatplotlibStyle(curve.style)

        x = curve.xs
        y = curve.ys

        [line] = axes.plot(
            x,
            y,
            mpl_style.line_color,
            linewidth=mpl_style.line_width,
            linestyle=mpl_style.line_style,
        )
        if mpl_style.fill_color or mpl_style.fill_pattern:
            [line] = plt.fill(
                x,
                y,
                mpl_style.fill_color,
                edgecolor=mpl_style.line_color,
                linewidth=mpl_style.line_width,
                hatch=mpl_style.fill_pattern,
            )

        if curve.style.arrow is not None:
            if curve.style.arrow in [Style.ArrowStyle.START, Style.ArrowStyle.DOUBLE]:
                x_s, y_s = x[0], y[0]
                dx_s, dy_s = x[1] - x[0], y[1] - y[0]
                self._plot_arrow(x_s, y_s, dx_s, dy_s, curve.style, axes)
            if curve.style.arrow in [Style.ArrowStyle.END, Style.ArrowStyle.DOUBLE]:
                x_e, y_e = x[-1], y[-1]
                dx_e, dy_e = x[-2] - x[-1], y[-2] - y[-1]
                self._plot_arrow(x_e, y_e, dx_e, dy_e, curve.style, axes)

        if mpl_style.shadow:
            # http://matplotlib.sourceforge.net/users/transforms_tutorial.html#using-offset-transforms-to-create-a-shadow-effect
            # shift the object over 2 points, and down 2 points
            dx, dy = mpl_style.shadow / 72.0, -mpl_style.shadow / 72.0
            offset = transforms.ScaledTranslation(dx, dy, fig.dpi_scale_trans)
            shadow_transform = axes.transData + offset
            # now plot the same data with our offset transform;
            # use the zorder to make sure we are below the line
            axes.plot(
                x,
                y,
                linewidth=mpl_style.line_width,
                color="gray",
                transform=shadow_transform,
                zorder=0.5 * line.get_zorder(),
            )

    def _plot_arrow(self, x, y, dx, dy, style: Style, axes: plt.Axes):
        """Draw arrow (dx,dy) at (x,y). `style` is '->', '<-' or '<->'."""
        mpl_style = MatplotlibStyle(style)
        if style.arrow in [Style.ArrowStyle.START, Style.ArrowStyle.DOUBLE]:
            axes.arrow(
                x,
                y,
                dx,
                dy,
                facecolor=mpl_style.line_color,
                edgecolor=mpl_style.line_color,
                linestyle=mpl_style.line_style,
                linewidth=mpl_style.line_width,
                # head_width=self.arrow_head_width,
                # head_width=0.1,
                # width=1,  # width of arrow body in coordinate scale
                length_includes_head=True,
                shape="full",
            )
        if style.arrow in [Style.ArrowStyle.END, Style.ArrowStyle.DOUBLE]:
            axes.arrow(
                x + dx,
                y + dy,
                -dx,
                -dy,
                facecolor=mpl_style.line_color,
                edgecolor=mpl_style.line_color,
                linewidth=mpl_style.line_width,
                head_width=0.1,
                # width=1,
                length_includes_head=True,
                shape="full",
            )



from typing import Any, Callable, Dict, TypeVar, Union

from pysketcher.point import Point
from pysketcher.shape import Shape, Stylable
from pysketcher.style import Style, TextStyle
from pysketcher.text import Text

T = TypeVar("type")


class Composition(Stylable):
    class CompositionStyle(Style):
        """Presents the Stylable contract for a Composition, setting the style of each object in the composition
        transparently """

        _composition: "Composition"

        def __init__(self, composition: "Composition"):
            super().__init__()
            self._composition = composition

        @property
        def line_style(self) -> Style.LineStyle:
            return super().line_style

        @line_style.setter
        def line_style(self, line_style: Style.LineStyle):
            for shape in self._composition:
                shape.style.line_style = line_style

        @property
        def line_width(self) -> float:
            return super().line_width

        @line_width.setter
        def line_width(self, line_width: float):
            for shape in self._composition:
                shape.style.line_width = line_width

        @property
        def line_color(self) -> Style.Color:
            return super().line_color

        @line_color.setter
        def line_color(self, line_color: Style.Color):
            for shape in self._composition:
                shape.style.line_color = line_color

        @property
        def fill_color(self) -> Style.Color:
            return super().fill_color

        @fill_color.setter
        def fill_color(self, fill_color: Style.Color):
            for shape in self._composition:
                shape.style.fill_color = fill_color

        @property
        def fill_pattern(self) -> Style.FillPattern:
            return super().fill_pattern

        @fill_pattern.setter
        def fill_pattern(self, fill_pattern: Style.Color):
            for shape in self._composition:
                shape.style.fill_pattern = fill_pattern

        @property
        def arrow(self) -> Style.ArrowStyle:
            return super().arrow

        @arrow.setter
        def arrow(self, arrow: Style.ArrowStyle):
            for shape in self._composition:
                shape.style.arrow = arrow

        @property
        def shadow(self) -> float:
            return super().shadow

        @shadow.setter
        def shadow(self, shadow: float):
            for shape in self._composition:
                shape.style.shadow = shadow

    _shapes: dict

    def __init__(self, shapes: dict):
        """shapes: list or dict of Shape objects."""
        super().__init__()
        self._shapes = shapes
        self._style = self.CompositionStyle(self)

    def __iter__(self):
        return self._shapes.values().__iter__()

    def add(self, key: str, shape: Shape) -> "Composition":
        """Provides a copy of the composition with the requisite shape added"""
        shapes = self._shapes.copy()
        shapes[key] = shape
        return Composition(shapes)

    def __setitem__(self, key: str, value: Shape) -> "Composition":
        return self.add(key, value)

    def __getitem__(self, name):
        return self._shapes[name]

    def _for_all_shapes(self, func: str, *args, **kwargs) -> "Composition":
        shapes = dict()
        for key, shape in self._shapes.items():
            shapes[key] = getattr(shape, func)(*args, **kwargs)
        return Composition(shapes)

    def rotate(self, angle: float, center: Point) -> "Composition":
        return self._for_all_shapes("rotate", angle, center)

    def translate(self, vec) -> "Composition":
        return self._for_all_shapes("translate", vec)

    def scale(self, factor) -> "Composition":
        return self._for_all_shapes("scale", factor)

    def apply(self, func: Callable[[Union[Shape, "Composition"]], T]) -> Dict[str, T]:
        ret_dict = {}
        for key, shape in self._shapes.items():
            ret_dict[key] = func(shape)
        return ret_dict


class ShapeWithText(Composition):
    def __init__(self, shape: Shape, text: Text):
        super().__init__({"text": text, "shape": shape})



from abc import ABC, abstractmethod

from pysketcher.point import Point
from pysketcher.style import Style


class Stylable(ABC):

    _style: Style

    @abstractmethod
    def __init__(self):
        self._style = Style()

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, style: Style):
        self._style = style

    def set_line_width(self, line_width: float) -> "Stylable":
        self.style.line_width = line_width
        return self

    def set_line_style(self, line_style: Style.LineStyle) -> "Stylable":
        self.style.line_style = line_style
        return self

    def set_line_color(self, line_color: Style.Color) -> "Stylable":
        self.style.line_color = line_color
        return self

    def set_fill_pattern(self, fill_pattern: Style.FillPattern) -> "Stylable":
        self.style.fill_pattern = fill_pattern
        return self

    def set_fill_color(self, fill_color: Style.Color) -> "Stylable":
        self.style.fill_color = fill_color
        return self

    def set_arrow(self, arrow: Style.ArrowStyle) -> "Stylable":
        self.style.arrow = arrow
        return self

    def set_shadow(self, shadow: float) -> "Stylable":
        self.style.shadow = shadow
        return self


class Shape(Stylable):
    """
    Superclass for drawing different geometric shapes.
    Subclasses define shapes, but drawing, rotation, translation,
    etc. are done in generic functions in this superclass.
    """

    @abstractmethod
    def __init__(self):
        super().__init__()
        self._name = None

    def __iter__(self):
        return [self]

    @abstractmethod
    def rotate(self, angle: float, center: Point) -> "Shape":
        raise NotImplementedError

    def translate(self, vec) -> "Shape":
        raise NotImplementedError

    def scale(self, factor) -> "Shape":
        raise NotImplementedError



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    pass

from pysketcher.angle import Angle
from pysketcher.arc import Arc, ArcWithText
from pysketcher.arrow import Arrow, DoubleArrow
from pysketcher.arrow_with_text import ArrowWithText
from pysketcher.axis import Axis
from pysketcher.beam import SimpleSupport
from pysketcher.circle import Circle
from pysketcher.composition import Composition, ShapeWithText
from pysketcher.curve import Curve
from pysketcher.dashpot import Dashpot
from pysketcher.distance_with_text import DistanceWithText
from pysketcher.drawable import Drawable
from pysketcher.figure import Figure
from pysketcher.force import Force, Gravity
from pysketcher.line import Line
from pysketcher.moment import Moment
from pysketcher.point import Point
from pysketcher.rectangle import Rectangle
from pysketcher.shape import Shape
from pysketcher.sketchy_func import (
    SketchyFunc1,
    SketchyFunc2,
    SketchyFunc3,
    SketchyFunc4,
)
from pysketcher.spline import Spline
from pysketcher.spring import Spring
from pysketcher.style import Style, TextStyle
from pysketcher.text import Text
from pysketcher.triangle import Triangle
from pysketcher.uniform_load import UniformLoad
from pysketcher.velocity_profile import VelocityProfile
from pysketcher.wall import Wall
from pysketcher.wheel import Wheel



from pysketcher.style import Style, TextStyle


class MatplotlibStyle:
    _style: Style
    LINE_STYLE_MAP = {
        Style.LineStyle.SOLID: "-",
        Style.LineStyle.DOTTED: ":",
        Style.LineStyle.DASHED: "--",
        Style.LineStyle.DASH_DOT: "-.",
    }
    FILL_PATTERN_MAP = {
        Style.FillPattern.CIRCLE: "O",
        Style.FillPattern.CROSS: "x",
        Style.FillPattern.DOT: ".",
        Style.FillPattern.HORIZONTAL: "-",
        Style.FillPattern.SQUARE: "+",
        Style.FillPattern.STAR: "*",
        Style.FillPattern.SMALL_CIRCLE: "o",
        Style.FillPattern.VERTICAL: "|",
        Style.FillPattern.UP_LEFT_TO_RIGHT: "//",
        Style.FillPattern.UP_RIGHT_TO_LEFT: "\\\\",
    }
    COLOR_MAP = {
        Style.Color.GREY: "grey",
        Style.Color.BLACK: "black",
        Style.Color.BLUE: "blue",
        Style.Color.BROWN: "brown",
        Style.Color.CYAN: "cyan",
        Style.Color.GREEN: "green",
        Style.Color.MAGENTA: "magenta",
        Style.Color.ORANGE: "orange",
        Style.Color.PURPLE: "purple",
        Style.Color.RED: "red",
        Style.Color.YELLOW: "yellow",
        Style.Color.WHITE: "white",
    }
    ARROW_MAP = {
        Style.ArrowStyle.DOUBLE: "<->",
        Style.ArrowStyle.START: "<-",
        Style.ArrowStyle.END: "->",
    }

    def __init__(self, style: Style):
        self._style = style

    @property
    def line_width(self) -> float:
        return self._style.line_width

    @property
    def line_style(self) -> str:
        return self.LINE_STYLE_MAP.get(self._style.line_style)

    @property
    def line_color(self):
        return self.COLOR_MAP.get(self._style.line_color)

    @property
    def fill_color(self):
        return self.COLOR_MAP.get(self._style.fill_color)

    @property
    def fill_pattern(self):
        return self.FILL_PATTERN_MAP.get(self._style.fill_pattern)

    @property
    def arrow(self):
        return self.ARROW_MAP.get(self._style.arrow)

    @property
    def shadow(self):
        return self._style.shadow

    def __str__(self):
        return (
            "line_style: %s, line_width: %s, line_color: %s,"
            " fill_pattern: %s, fill_color: %s, arrow: %s shadow: %s"
            % (
                self.line_style,
                self.line_width,
                self.line_color,
                self.fill_pattern,
                self.fill_color,
                self.arrow,
                self.shadow,
            )
        )


class MatplotlibTextStyle(MatplotlibStyle):
    FONT_FAMILY_MAP = {
        TextStyle.FontFamily.SERIF: "serif",
        TextStyle.FontFamily.SANS: "sans-serif",
        TextStyle.FontFamily.MONO: "monospace",
    }

    ALIGNMENT_MAP = {
        TextStyle.Alignment.LEFT: "left",
        TextStyle.Alignment.RIGHT: "right",
        TextStyle.Alignment.CENTER: "center",
    }

    _style: TextStyle

    def __init__(self, text_style: TextStyle):
        super().__init__(text_style)

    @property
    def font_size(self) -> float:
        return self._style.font_size

    @property
    def font_family(self) -> str:
        return self.FONT_FAMILY_MAP.get(self._style.font_family)

    @property
    def alignment(self) -> str:
        return self.ALIGNMENT_MAP.get(self._style.alignment)



from pysketcher.line import Line
from pysketcher.point import Point
from pysketcher.style import Style


class Arrow(Line):
    """Draw an arrow as Line with arrow pointing towards end."""

    def __init__(self, start: Point, end: Point):
        super().__init__(start, end)
        self.style.arrow = Style.ArrowStyle.END


class DoubleArrow(Line):
    """Draw an arrow as Line with arrow pointing towards start and end."""

    def __init__(self, start: Point, end: Point):
        super().__init__(start, end)
        self.style.arrow = Style.ArrowStyle.DOUBLE



import numpy as np

from pysketcher.circle import Circle
from pysketcher.composition.composition import Composition
from pysketcher.line import Line
from pysketcher.point import Point


class Wheel:
    _center: Point
    _radius: float
    _inner_radius: float
    _nlines: int

    def __init__(
        self, center: Point, radius: float, inner_radius: float = None, nlines: int = 10
    ):
        self._center = center
        self._radius = radius
        self._inner_radius = inner_radius
        self._nlines = nlines

        if inner_radius is None:
            self._inner_radius = radius / 5.0

        outer = Circle(center, radius)
        inner = Circle(center, inner_radius)
        lines = []
        # Draw nlines+1 since the first and last coincide
        # (then nlines lines will be visible)
        t = np.linspace(0, 2 * np.pi, nlines + 1)

        xinner = self._center.x + self._inner_radius * np.cos(t)
        yinner = self._center.y + self._inner_radius * np.sin(t)
        xouter = self._center.x + self._radius * np.cos(t)
        youter = self._center.y + self._radius * np.sin(t)
        lines = [
            Line(Point(xi, yi), Point(xo, yo))
            for xi, yi, xo, yo in zip(xinner, yinner, xouter, youter)
        ]
        self.shapes = {
            "inner": inner,
            "outer": outer,
            "spokes": Composition({"spoke%d" % i: lines[i] for i in range(len(lines))}),
        }



from pysketcher.curve import Curve
from pysketcher.point import Point
from pysketcher.text import Text


class Triangle(Curve):
    """
    Triangle defined by its three vertices p1, p2, and p3.

    Recorded geometric features:

    ==================== =============================================
    Attribute            Description
    ==================== =============================================
    p1, p2, p3           Corners as given to the constructor.
    ==================== =============================================

    """

    def __init__(self, p1: Point, p2: Point, p3: Point):
        self._p1 = p1
        self._p2 = p2
        self._p3 = p3

        # Dimensions
        self.dimensions = {
            "p1": Text("p1", p1),
            "p2": Text("p2", p2),
            "p3": Text("p3", p3),
        }
        super().__init__([p1, p2, p3, p1])

    def geometric_features(self):
        return {"p1": self._p1, "p2": self._p2, "p3": self._p3}

    def rotate(self, angle: float, center: Point):
        return Triangle(
            self._p1.rotate(angle, center),
            self._p2.rotate(angle, center),
            self._p3.rotate(angle, center),
        )



from pysketcher.arrow_with_text import ArrowWithText
from pysketcher.point import Point
from pysketcher.style import Style


class Force(ArrowWithText):
    """Another name for `ArrowWithText` to make code easier to read"""

    pass


class Gravity(Force):
    """Downward-pointing gravity arrow with the symbol g."""

    def __init__(
        self,
        start,
        length,
        text_position: ArrowWithText.TextPosition = ArrowWithText.TextPosition.START,
        text="$g$",
    ):
        Force.__init__(
            self, text, start, start + Point(0, -length), text_position=text_position
        )
        self.style.line_color = Style.Color.BLACK



from typing import List

import numpy as np
from scipy.interpolate import UnivariateSpline

from pysketcher.curve import Curve
from pysketcher.point import Point


class Spline(Curve):

    _input_points: List[Point]
    _smooth: UnivariateSpline

    # Note: UnivariateSpline interpolation may not work if
    # the x[i] points are far from uniformly spaced
    def __init__(self, points: List[Point], degree: int = 3, resolution: int = 501):
        self._input_points = points
        self._smooth = UnivariateSpline(
            [p.x for p in points], [p.y for p in points], s=0, k=degree
        )
        x_coordinates = np.linspace(points[0].x, points[-1].x, resolution)
        y_coordinates = self._smooth(x_coordinates)
        smooth_points = [Point(p[0], p[1]) for p in zip(x_coordinates, y_coordinates)]
        super().__init__(smooth_points)

    def geometric_features(self):
        return {
            "start": self.points[0],
            "end": self.points[-1],
            "interval": [self.points[0].x, self.points[-1].x],
        }

    def __call__(self, x):
        return self._smooth(x)

    # Can easily find the derivative and the integral as
    # self.smooth.derivative(n=1) and self.smooth.antiderivative()
